"""Types used throughout the project"""

import pathlib
from typing import Any, Union

PathType = Union[str, pathlib.Path]
IntArrayType = Any  # np.ndarray[int]
FloatArrayType = Any  # np.ndarray[float]
